#include "main.h"

// -----------------------------------------------------------------------------
// ОБЩИЕ МЕТОДЫ
// -----------------------------------------------------------------------------

Main::Main(int w, int h, int scale, int fps) {
    create(w, h, scale, "SDL2 Window", fps);
}

Main::~Main() {
    destroy();
}

// Создать новое окно
int Main::create(int w, int h, int scale, const char* name = "SDL2", int fps = 50) {

    unsigned format = SDL_PIXELFORMAT_BGRA32;

    _scale  = scale;
    _width  = w;
    _height = h;

    width  = w * scale;
    height = h * scale;

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO)) {
        exit(1);
    }

    SDL_ClearError();

    // Создать окно
    sdl_window = SDL_CreateWindow(name, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, width, height, SDL_WINDOW_SHOWN);
    if (sdl_window == NULL) exit(1);

    // Создать отрисовщик текстур
    sdl_renderer = SDL_CreateRenderer(sdl_window, -1, SDL_RENDERER_PRESENTVSYNC);
    if (!sdl_renderer) exit(1);

    // Формат пикселей
    sdl_pixel_format = SDL_AllocFormat(format);
    if (!sdl_pixel_format) exit(1);

    //  Создать текстуру в памяти
    sdl_screen_texture = SDL_CreateTexture(sdl_renderer, format, SDL_TEXTUREACCESS_STREAMING, width, height);
    if (!sdl_screen_texture) exit(1);

    // Смешивания текстур нет
    SDL_SetTextureBlendMode(sdl_screen_texture, SDL_BLENDMODE_NONE);

    // Буфер экрана в памяти
    screen_buffer    = (Uint32*)malloc(width * height * sizeof(Uint32));

    // Настройка FPS
    frame_length     = 1000 / (fps ? fps : 1);
    frame_prev_ticks = SDL_GetTicks();

    return 0;
}

// Ожидание событий
int Main::event() {

    for (;;) {

        Uint32 ticks = SDL_GetTicks();

        // Ожидать наступления события
        while (SDL_PollEvent(& evt)) {

            switch (evt.type) {

                // Выход из программы по нажатии "крестика"
                case SDL_QUIT: {
                    return 0;
                }

                // https://wiki.machinesdl.org/SDL_Scancode

                // Нажатие на клавишу
                case SDL_KEYDOWN: {

                    kb = evt.key.keysym.scancode;
                    return EvtKbDown;
                }

                // Клавиша отпущена
                case SDL_KEYUP: {

                    kb = evt.key.keysym.scancode;
                    return EvtKbUp;
                }

                // Движение мыши
                case SDL_MOUSEMOTION: {

                    mx = evt.motion.x;
                    my = evt.motion.y;
                    return EvtMsMotion;
                }

                // Движение мыши
                case SDL_MOUSEBUTTONDOWN:
                case SDL_MOUSEBUTTONUP: {

                    // SDL_BUTTON_LEFT | SDL_BUTTON_MIDDLE | SDL_BUTTON_RIGHT
                    mb = evt.button.button; 

                    // SDL_PRESSED | SDL_RELEASED
                    ms = evt.button.state;  

                    return EvtMsButton;
                }

                // Все другие события
                default: {
                    return EvtOther;
                }
            }
        }

        // Истечение таймаута: обновление экрана
        if (ticks - frame_prev_ticks >= frame_length) {

            frame_prev_ticks = ticks;
            update();
            return EvtRedraw;
        }

        SDL_Delay(1);
    }
}

// Обновить окно
void Main::update() {

    SDL_Rect dstRect;

    dstRect.x = 0;
    dstRect.y = 0;
    dstRect.w = width;
    dstRect.h = height;

    SDL_UpdateTexture       (sdl_screen_texture, NULL, screen_buffer, width * sizeof(Uint32));
    SDL_SetRenderDrawColor  (sdl_renderer, 0, 0, 0, 0);
    SDL_RenderClear         (sdl_renderer);
    SDL_RenderCopy          (sdl_renderer, sdl_screen_texture, NULL, &dstRect);
    SDL_RenderPresent       (sdl_renderer);
}

// Уничтожение окна
void Main::destroy() {

    if (sdl_screen_texture) {
        SDL_DestroyTexture(sdl_screen_texture);
        sdl_screen_texture = NULL;
    }

    if (sdl_pixel_format) {
        SDL_FreeFormat(sdl_pixel_format);
        sdl_pixel_format = NULL;
    }

    if (sdl_renderer) {
        SDL_DestroyRenderer(sdl_renderer);
        sdl_renderer = NULL;
    }

    free(screen_buffer);

    SDL_DestroyWindow(sdl_window);
    SDL_Quit();
}

// -----------------------------------------------------------------------------
// ФУНКЦИИ РИСОВАНИЯ
// -----------------------------------------------------------------------------

// Установка точки
void Main::pset(int x, int y, Uint32 cl) {

    if (x < 0 || y < 0 || x >= _width || y >= _height)
        return;

    if (_scale == 1) {
        screen_buffer[y*width + x] = cl;
    } else {
        for (int i = 0; i < _scale; i++)
        for (int j = 0; j < _scale; j++)
            screen_buffer[(_scale*y+i)*width + (_scale*x + j)] = cl;
    }
}

// Рисовать линию
void Main::line(int x1, int y1, int x2, int y2, Uint32 color) {

    // Инициализация смещений
    int signx  = x1 < x2 ? 1 : -1;
    int signy  = y1 < y2 ? 1 : -1;
    int deltax = x2 > x1 ? x2 - x1 : x1 - x2;
    int deltay = y2 > y1 ? y2 - y1 : y1 - y2;
    int error  = deltax - deltay;
    int error2;

    // Если линия - это точка
    pset(x2, y2, color);

    // Перебирать до конца
    while ((x1 != x2) || (y1 != y2)) {

        pset(x1, y1, color);
        error2 = 2 * error;

        // Коррекция по X
        if (error2 > -deltay) {
            error -= deltay;
            x1 += signx;
        }

        // Коррекция по Y
        if (error2 < deltax) {
            error += deltax;
            y1 += signy;
        }
    }
}

// Рисование сплошного блока
void Main::block(int x, int y, int w, int h, Uint32 color) {

    for (int i = y; i < y + h; i++) {

        if (i >= _height) break;
        for (int j = x; j < x + w; j++) {

            if (j >= _width) break;
            pset(j, i, color);
        }
    }
}

// Рисовать окружность
void Main::circle(int xc, int yc, int radius, int color) {

    int x = 0,
        y = radius,
        d = 3 - 2*y;

    while (x <= y) {

        // Верхний и нижний сектор
        pset(xc - x, yc + y, color);
        pset(xc + x, yc + y, color);
        pset(xc - x, yc - y, color);
        pset(xc + x, yc - y, color);

        // Левый и правый сектор
        pset(xc - y, yc + x, color);
        pset(xc + y, yc + x, color);
        pset(xc - y, yc - x, color);
        pset(xc + y, yc - x, color);

        d += (4*x + 6);
        if (d >= 0) {
            d += 4*(1 - y);
            y--;
        }

        x++;
    }
}

// Рисовать закрашенную окружность
void Main::circlef(int xc, int yc, int radius, int color) {

    int x = 0,
        y = radius,
        d = 3 - 2*y;

    while (x <= y) {

        // Верхний и нижний сектор
        line(xc - x, yc + y, xc + x, yc + y, color);
        line(xc - x, yc - y, xc + x, yc - y, color);

        // Левый и правый сектор
        line(xc - y, yc + x, xc + y, yc + x, color);
        line(xc - y, yc - x, xc + y, yc - x, color);

        d += (4*x + 6);
        if (d >= 0) {
            d += 4*(1 - y);
            y--;
        }

        x++;
    }
}